/*
CIFFOLD 0.5.4 Pre-Release
by Kostadin Mitev, Georgi Todorov and Herbert J. Bernstein

Copyright (C) Kostadin Mitev 2005

Work funded in part by the International Union of Crystallography under
a grant to Dowling College.

This software is covered by the GNU General Public License 
<see COPYING>.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

 */

//folding and unfolding of CIF files 

#include<stdlib.h>
#include<stdio.h>
#include<string>
#include<sstream>
#include<ctype.h>
#include<iostream>
#include"FUCIF.h"
#include"Globals.h"
#ifndef isblank
#define isblank(c) (((c)=='\t')||((c)==' '))?1:0
#endif

#define MAXLNG 80
#define TABLEN 8

#define DEBUG 0

///////////LOCAL GLOBALS////////////////////////
long int LINECOUNT = 0;			//counts the lines read
bool EXTRAHASHMARK = false;		//Shows whether there should be an extra hash mark 
					//inserted used when folding only. If there is a 
					//folded comment and it finishes with a backslash 
					//and after it there is another comment there 
					//should be an extrahashmark inserted to
					// avoid merging the two comments when unfolding the file
int FRTEXTCNT = -1;			//Counts the number of textfields
int NLOOP = -1;				//Counts the number of loops
char *VERSIONS[4] = {"\0", "1.0", "1.1" };	//Holds the valid versions used in insertvVersion
vector <long>::iterator FChIterator;		
long START = 0;					 //used only when parts of the file should be processed
long END = 0;					 //used only when parts of the file should be processed
bool PRESERVE;
FILE *MAPFILE;
char *LINEMAP = new char [BUFSIZE + 1];		//holds the map of the current line the map consists 
						//of the following: /tn where n indicates
						//the number of tabs to be inserted /d means that 
						//there should be data /n where n is a digid 
						//showing the number of blanks and v for version
bool MAPON = false; 		//show whether the file should be mapped according to the map table in file MAPFILE
int LMCOUNT = 0;
bool versionInserted = false;	//if folding and we have inserted a version then set this option which wiil will indicate that the map of the file should be adjusted
bool afterversion = false; 	//If there is something after the version on the same line then this is set to true to indicate that another d has to be added to the map

void  reader()
{
	char * curLine = new char[BUFSIZE + 1];
		//(char *)malloc(2052*sizeof(char));
	char state = 0;    //2-for a comment 1- for a textfield, 3 for converted text field(long strings), 4 for egular data that needs to 			     //be converted.
	int pos = 0, ii = 0;
	int left = 0; 			//Left space on line
	int strcount = 0;
	int stlen = 0;
	int curBlanks = 0; 		//shows the blanks between items on one line
	int freesp = 0;			//shows the free spaces left until we reach MAXLENGTH
	int terse = 0;
	int Lblanks =-1;		//Number of leading blanks
	int Tblanks = -1;		//Number of trailing blanks
	int insBl = 0;			//Holds the numberof inserted leading blanks
	char Delim[] = " ";
	char *token = new char [81];
	char *tmpstr1;
	char *tmpstr2;
	int ch = 0;			//0-do not preserve leading blanks, 1-preserve leading blanks
	bool formatPrevLine = true;	
	FChIterator = FileChunks.begin();
	//if the option for mapping is set then open the map file
	if (OPTIONS[13]){
		MAPFILE = fopen((R.getMapfilename()), "r");
		if(!MAPFILE){
			if(OPTIONS[1]){
				cerr<<"ERROR occured while trying to open the map-file"<<endl;
		       		cerr<<"Your file will be formatted with the default options"<<endl;
			}
			else{
				cerr<<"There is a problem with creating the MAP and it will not be created!!!"<<endl;
			}
		}
		else if(!OPTIONS[1]){
			MAPON = true;
		}
		
	}
	left = MAXLENGTH;
//////////////////////////////////////ASSIGN THE OPTIONS//////////////////////////////////////////////
	if(OPTIONS[2] == 1){
		terse = 1;

	}
	ch = OPTIONS[4];		//if 0 do not keep leading blanks
	fgets(curLine, BUFSIZE, INF);
	if( (MAPON) && (!OPTIONS[1]) ){
		fgets(LINEMAP, BUFSIZE, MAPFILE);
		rmtblanks(LINEMAP);
		LMCOUNT = 0;
	}
	if(OPTIONS[0] != 0){
		insertVersion(OPTIONS[0], curLine, pos, left);
		pos += rmfblanks(curLine, pos);
		if(!curLine[pos]){
			LINECOUNT++;
			if(MAPON){
				checkNewLineMAP(left);
				//if the linemap has been used and we are in unfolding then take the next one
				if( (!LINEMAP[LMCOUNT]) && (!OPTIONS[1]) ){
//	fputc('\n', OUTF);
//	left = MAXLENGTH;
					fgets(LINEMAP, BUFSIZE, MAPFILE);
					rmtblanks(LINEMAP);
					LMCOUNT = 0;
				}
			}
			fgets(curLine, BUFSIZE, INF);
			pos = 0;
		}
	}
	
////////////////READING THE FILE///////////////////////////////////////	
	do{
	  formatPrevLine = formatLine();
	  LINECOUNT++;
//insert an extra hashmark if necessary
	  if(EXTRAHASHMARK){
		  Lblanks = rmfblanks(curLine, pos);
	      if (curLine[Lblanks] == '#'){
		      fputs("#\n", OUTF);
		      left = MAXLENGTH;
	      }
	      EXTRAHASHMARK = false;
 	  }
/////////DETERMINE WHETHER THE STRING SHOULD BE PROCESSED////////////////	  
    //if a special comment do not process	  
	  if(!strncmp(curLine, "#_#", 3)){
  	      if(!MAPON){
		  if(left < MAXLENGTH){
		      	fputc('\n', OUTF);
		      	left = MAXLENGTH;
		  }
	      }
	      else{
 			rmtblanks(curLine);			      
		      	insertMapBlanks(left, 1);
			left -= strlen(curLine);	//since the trailing blanks have been removed no new line has been outputd so adjust left
			LMCOUNT++;
	      }
	      fputs(curLine, OUTF);
   	      continue;
	  }
	  if(OPTIONS[1]){
	  	foldline(curLine, pos, left);
	  }
	  else{
		 //if we are using the map and format only file chuncks is selected
		if((MAPON)&&(!OPTIONS[6])){
			//if a the previous line was formated and the current should not be make sure it is on a new line.
			if((formatPrevLine)&&(!formatLine())){
				if(left < MAXLENGTH){
					
					fputc('\n', OUTF);
					left = MAXLENGTH;
				}
			}
		}
	  	unfoldline(curLine, pos, left);
	  }
	  pos = 0;
	}while(fgets(curLine, BUFSIZE, INF));	
       //if the errormessages or the warnings should be printed 
	if((OPTIONS[11]) ||(OPTIONS[12])){
		R.printErr(OUTF, OPTIONS[11], OPTIONS[12]);
		
	}
	//Prints the error messages and the warnings to stderr
	R.printErr();
///////////THE NEW OPTIONS GENERATE MAP HERE//////////
//if generate a map and folding the output concatenate the map to the file
    if(OPTIONS[13] == 1){ 
	  left = MAXLENGTH;			//the lines of the map should be no longer than MAXLENGTH
	  int delimlen = 4;			//The lelngth of the opening dilimeter of the map #_M#
	  bool isfirst = true;			//marks the first outputed line of the  map
	  pos = 0;				

	    if( (OPTIONS[1] == 1) && (MAPFILE) ){
       //if we have inserted a version in the file then adjust the map by adding a new data field at the begining		  
	   if(versionInserted){
	   	fputs("#_M#d", OUTF);
		left -= delimlen +1;
		isfirst = false;
	   }
	   //if was something after the fileversion inentifyer and was put on a new line insert indecate so on the mapfile
	   else if(afterversion){  
		 //read one line which will hold the map of the version 
		fgets(curLine, BUFSIZE, MAPFILE);
		//the version has to always be at the beginning of the file so ignore the maping which may be wrong!	
	     	fputs("#_M#dnd", OUTF);
//fputs("#_M#d\n", OUTF);
		left -= delimlen +3;
		isfirst = false;
	   } 
	   if(isfirst){
	   	fgets(curLine, BUFSIZE, MAPFILE);
		rmtblanks(curLine);
		fputs("#_M#", OUTF);
		left -= delimlen;
		while(curLine[pos]){
			if(left <= 0){
				fputs("\n#_M#", OUTF);
				left = MAXLENGTH - delimlen;
			}
			fputc(curLine[pos], OUTF);
			left--;
			pos++;
		}
	   } 
	   while(fgets(curLine, BUFSIZE, MAPFILE))
	   {
		rmtblanks(curLine);
		pos =0;
		if(left <= 0){
			fputs("\n#_M#", OUTF);
			left = MAXLENGTH - delimlen;
		}
		fputc('n', OUTF);
		left--;
		while(curLine[pos]){
			if(left <= 0){
				fputs("\n#_M#", OUTF);
				left = MAXLENGTH - delimlen;
			}
			fputc(curLine[pos], OUTF);
			left--;
			pos++;
		}
//fputs("#_M#", OUTF);   
//fputs(curLine, OUTF);
    	   }
	  }

	if(MAPFILE){
		fclose(MAPFILE);
	  	remove(R.getMapfilename());
	}
    }
/////////////////DELETE ALLOCATED MEMORY//////////////////////
   delete [] curLine;
   delete [] token;
   delete [] LINEMAP;
   delete [] R.getMapfilename();
   token = NULL;
   curLine = NULL;
}
//Inserts as meny blanks as possible
int insertBlanks(int curBl, int freeS)
{
    int g = 0;    
    if(curBl < freeS){
      freeS = curBl;
    }
    
    for (g = 0;g < freeS;g++)
    {
      fputc(' ', OUTF);
    }
    return freeS;
}
int insertMapBlanks(int &left, int insert)
{	
	//The insert indicates if the blanks should be inserted or just skiped
	//Find the type and number of blanks before the token and output them
	int howmany = 0;		//how many blanks or tabs have to be outputed
	int totalSingleBlanks = 0;	//what is the total number of single blanks outputed


	checkNewLineMAP(left);
/*	
	//Check if there is something on the mapline and if not read the next one
	if(!LINEMAP[LMCOUNT]){
	    do{	
		if(insert == 1){
			fputc('\n', OUTF);
			left = MAXLENGTH; 
		}
		else if(insert == 2){
			insert = 1;	//make it one so if there is another new line to be inserted it is not being skiped 
		}
		if(fgets(LINEMAP, BUFSIZE, MAPFILE)){
			rmtblanks(LINEMAP);
			LMCOUNT = 0;
			checkNewLineMAP(left);
		}
		else{
			MAPON = false;
//fprintf(stderr, "The MAP file was exhausted at line %d. Continuing with the defaults.\n", COUNTLINE);
		}
	    } while( (!LINEMAP[LMCOUNT]) && (MAPON));
		
	}
	if(MAPON){
	*/
	do{
		rmtblanks(LINEMAP);	 //remove the trailing blanks
		checkNewLineMAP(left);	 //insert new line if necessary NOTE: affects the global 'left' too
		//start inserting the blanks
		while(LINEMAP[LMCOUNT])
		{
			if(LINEMAP[LMCOUNT] == 't'){
				LMCOUNT++;
				if(convertToInt(howmany)){
					for(int y = 0; y < howmany; y++)
					{
					    if(insert){
						fputc('\t', OUTF);
					    }
					    totalSingleBlanks += TABLEN;	
					}
				}
				else{
					fprintf(stderr, "Invalid MAP character. Continuing with the defaults from line %d.\n", LINECOUNT);
					MAPON = false;
				}
			}
			else if(LINEMAP[LMCOUNT] == 's'){
				LMCOUNT++;
				if(convertToInt(howmany)){
					for(int y = 0; y < howmany; y++)
					{
						totalSingleBlanks++;
						if(insert){
							fputc(' ', OUTF);
						}
					}
				}
				else{
					fprintf(stderr, "Invalid MAP character. Continuing with the defaults from line %d.\n", LINECOUNT);
				       	MAPON = false;
				}
			}
			else{
				return totalSingleBlanks;
			}
		}
		LMCOUNT = 0;
	}while(fgets(LINEMAP, BUFSIZE, MAPFILE));
	if(!MAPON){
	     if(left < MAXLENGTH)  {   
		fputc('\n', OUTF);
		left = MAXLENGTH;
	     }
	     return 0;
	}
	return totalSingleBlanks;
	
}
bool  convertToInt(int &i)
{
	if(!LINEMAP[LMCOUNT]){
		fgets(LINEMAP, BUFSIZE, MAPFILE);
		LMCOUNT = 0;
		rmtblanks(LINEMAP);
	}
	//copy the digits to a string, convert the string to an int and return true if successful false otherwise
	string digits;
      //  digits	+= LINEMAP[LMCOUNT];
	
//	LMCOUNT++;
	while(isdigit(LINEMAP[LMCOUNT]))
	{
		digits += LINEMAP[LMCOUNT];
		LMCOUNT++;
		if(!LINEMAP[LMCOUNT]){
			fgets(LINEMAP, BUFSIZE, MAPFILE);
			LMCOUNT = 0;
		}
	}
	istringstream sStream(digits);

	if (sStream>>i){
		return true;
	}
	return false;	    
	
}


//Remove the leading blanks
int rmfblanks(char * curLine, int pos)
{
   int bl;
   //note isblank does not count the new line character 
       for ( bl = 0;( (curLine[pos]) && (isblank(curLine[pos])) ); pos++, bl++) ;
       return bl;
}

//Remove the trailing blanks
int rmtblanks(char * curLine)
{
	int bl;
	int size = strlen(curLine) - 1;
	for ( bl = 0;( (curLine[size]) && (isspace(curLine[size])) ); size--, bl++)
	{
		curLine[size] = '\0';
	}
	return bl;
}


//Remove the trailing blanks in a text line (preserving the '\n'
int rmttblanks(char * curLine)
{
	int bl;
	char tc;
	int size = strlen(curLine) - 1;
	tc = curLine[size];
	for ( bl = 0;( (curLine[size]) && (isspace(curLine[size])) ); size--, bl++)
	{
		curLine[size] = '\0';
	}
	if (tc=='\n') 
	{
		curLine[++size] = tc;
		bl--;
		curLine[size+1] = '\0';
	}
	return bl;
}


//Finds the length of a token
int findTokLen(char * curLine, int pos, int state)
{
	int tmp = pos;
	char terminator;
	
	if(state == 0){
		while( (!isspace(curLine[tmp])) && (curLine[tmp] != '\0') )
		{
			tmp++;
		}
	}
	else{
		(state == 3)?terminator = '\'':terminator = '"';
		while(curLine[tmp])
		{
			if( (curLine[tmp] == terminator) && ((!curLine[tmp + 1]) ||(isspace(curLine[tmp + 1]))) ){
				tmp++;
				break;
			}
			tmp++;
		}
	}
	return tmp - pos;
}

void  outputTextField(char * curLine, int &pos, int &left){
	if(DEBUG){
		fputs("-----STARToutputTextField------", OUTF);
     	}
	pos = 0;
	if(left < MAXLENGTH){
		fputc('\n', OUTF);
		left = MAXLENGTH;	
	}

	do
	{
		fputs(curLine, OUTF);
		LINECOUNT++;
	}while ( fgets(curLine, BUFSIZE, INF) 
	          && ((curLine[0] != ';') 
	               || !isspace(curLine[1]) 
	             ) 
	       );

	if(curLine[pos] == ';'){
		rmtblanks(curLine);
		fputc(';', OUTF);
		pos = 1;
		left = MAXLENGTH -1;	//there is only a semicolon on the line
	}
	 if(DEBUG){
	      fputs("-----ENDoutputTextField------", OUTF);
    	 }
}

int insertVersion(int vers, char *curLine, int &pos, int &left)
{
	/*Looks for an already specified version by looking at curLine,  if it finds one changes it to the correct one and outputs it
	 */
	char *version = new char[15];
	int blanks = 0;
	bool newline = false;
	
	blanks = rmfblanks(curLine, pos);
		
	if(!strncasecmp("#/#cif_", (curLine + pos + blanks), 7)) {
		//get rid of the new line at the end of line
		if(curLine[strlen(curLine) -1] == '\n'){
			curLine[strlen(curLine) -1] = '\0';
		}
		pos += blanks;
		//there is a version specified copy its delimiter
		strncpy(version, (curLine + pos), 7);
		version[7] = '\0';
		//compute the correct place from where porcessing should continue(the end of the specufyed version in curLine) and if mapon move to the next token
		while((curLine[pos]) && (!isspace(curLine[pos])))
		{
			pos++;
		}
		//if thre is a version already specifyed then there is a map for it so advance toward the next token`s map
		if(MAPON){
			insertMapBlanks(left, 0);
			LMCOUNT++;
		}
		else{
			newline = true;
		}
	}
	else{
		//if a map is generated and we are in folding then set the boolean versionInserted
		if( (OPTIONS[13]) && (OPTIONS[1])){
			versionInserted = true;	
		}
		strcpy(version, "#/#cif_");
		newline = true;
	}
	strcat(version, VERSIONS[vers]);
	fputs(version, OUTF);
	if(newline){
		fputs("\n", OUTF);
	}
	//If there was a version recognized and there is something after the version identifyer then output it on a new line prefixed with a comment identifyer

	if(pos > 0){
		blanks = rmfblanks(curLine, pos);
		pos += blanks;
		
		if(curLine[pos]){
				if(!newline){
					fputc('\n', OUTF);
				}
				afterversion = true;
//!!!!TO DO Should check the length of the line and if longer thatn allowed fold it		
				fputc('#', OUTF);
				fputs((curLine + pos), OUTF);
				pos = strlen(curLine);
				if(!MAPON){	
				   fputc('\n', OUTF);
				}
		}
	}
	left = MAXLENGTH;
	delete [] version;
	version = NULL;
}

int foldline(char *curLine, int &pos, int &left)
{
	if(DEBUG){
		fputs("-----STARTFORMATLINE------", OUTF);	
	}
	/*Tokenizes curLine starting from position "pos" and formats its content*/
	int state = 0; 			//shows the type of the token
	int toklen = 0;			//length of token
	int leadBlanks = 0;		//The number of leading blanks outputed
	int freesp = 0;			//the freespace betwoeen the tokens
	int curBlanks = 0;		//the balnks between two tokens
	bool formatline = formatLine();
	
    //if this is the beginning of the line deal with the leading blanks if any	
	if(pos == 0){
	     if( (!formatline) && (left < MAXLENGTH) ){
	     	fputc('\n', OUTF);
		left = MAXLENGTH;
	     }
		 if (curLine[0] != ';') {
	     	rmtblanks(curLine);
	     }
	     else rmttblanks(curLine);
//	     else if(curLine[strlen(curLine) -1] == '\n'){
//	     	curLine[strlen(curLine) - 1] = '\0';
//	     }
		leadBlanks = rmfblanks(curLine, pos);
		pos = leadBlanks;
		if(EXTRAHASHMARK){
		     if(curLine[leadBlanks] == '#'){
			fputs("#\n", OUTF);
			left = MAXLENGTH;
		     }
		     EXTRAHASHMARK = false;
		}
		pos = leadBlanks;
		if(OPTIONS[4] == 1){			//if the leading  blanks should be kept output as many as possible
			leadBlanks = insertBlanks(leadBlanks, left);
			left -= leadBlanks;
		}
		else{
			leadBlanks = 0;
		}
	}
	else{
		curBlanks = rmfblanks(curLine, pos);
		rmtblanks(curLine);
	}
      //if there is something on the line and terse is selected at least one blank character has to be outputed before the next token	
	if((OPTIONS[2]) && (left < MAXLENGTH) && (!MAPON)){
		curBlanks = 1;
	}
     //there is nothing on the line and MAP is no on and terse is not selected output a new line
//        if((!curLine[pos]) && (!MAPON) && (!OPTIONS[2]) ){
//		fputc('\n', OUTF);
//		left = MAXLENGTH;	
//	}
	while(curLine[pos])
	{
		switch(curLine[pos])
		{
	//////////////TEXTFIELDS//////////////	
			case(';'):
			{
				//if the semicolumn is at position 0 then it is a textfield
				if(pos == 0){
					state = 1;
				//if the text field has lines longer than allowed it have to be folded else or 
				//  if only comments should be formated just output it	
					FRTEXTCNT++;
					if((!FORMATTEXT[FRTEXTCNT]) || (OPTIONS[7]) || (!formatline) ){
						outputTextField(curLine, pos, left);
						if(left < MAXLENGTH){
							fputc('\n', OUTF);
							left = MAXLENGTH;
						}
					}
					else{
						folding(state, 0, curLine, left, pos);	
					}
					formatline = formatLine();
				}
				else{
					state = 0;
				}
			  break;
			}
			case('#'):
			{
				state = 2;
			//if format everything except comments just output the comment wherever it is	
				if( (OPTIONS[8] == 1) || (!formatline) ){
					freesp = strlen(curLine);
					left -= insertBlanks(curBlanks, freesp);
					fputs( (curLine + pos), OUTF);
					pos = strlen(curLine); 
					if(curLine[strlen(curLine) -1] != '\n'){
						fputc('\n', OUTF);
						left = MAXLENGTH;
					}
					break;
				}
				//if this is just a hash mark and nothing after it put it on a new line 
				//It is usually used as a delimiter between blocks of the file to improve file reading
				if(!curLine[pos + 1]){
					if(left < MAXLENGTH){
						fputc('\n', OUTF);
					}
						fputc(curLine[pos], OUTF);
						pos++;
						left = MAXLENGTH -1;
				}
				else{
					toklen = strlen(curLine+pos); //-1
				//if there is something on the line compute the freespace between the tokens and output as many balanks as possible
				// if not terse 
					if(left < MAXLENGTH - leadBlanks){
						freesp = left - toklen;
						if((freesp < 1) && (left < 3)){
						     //if there is enough space for one space char the opening delimiter of a folded comment and one char of the comment call folding	
							fputc('\n', OUTF);
							left = MAXLENGTH;
							freesp = 0; 
						}
				     	//If terse is selected and freesp is more than 1 	
						else if((OPTIONS[2] == 1) || (freesp < 0)){
							freesp = 1;
						}
						left -= insertBlanks(curBlanks, freesp);
					}	
					if(toklen > left){
				//if the last character is a backslash and there is another comment after it then we have to insert an extra hashmark between the two
						if(curLine[ pos + toklen -1] == '\\'){	
							strcat(curLine, "\\");  //add an extra hashmark at the end so when unfolding we preserve the comment`s backslash
							EXTRAHASHMARK = true;
						}
						folding(state, 1, curLine, left, pos);
					}
					else{
						outputToken(curLine, pos, pos + toklen);
						left -= toklen;//take out -1
					}
				}
				//if the formatin is terse put a new line
				if(OPTIONS[2]){
					fputc('\n', OUTF);
					left = MAXLENGTH;
				}	
			  break;
			}
			case('\''):
			{
				state = 3;
			//go to the next case satatement since the quoted strings are handled in the same way
			}
			case('\"'):
			{
				if(state != 3){
					state = 4;
				}
				toklen = findTokLen(curLine, pos + 1, state);
				toklen++;	//one more for the opening delemeter
				if( (OPTIONS[7] == 1) || (!formatline) ){
					freesp = strlen(curLine);	//make freesp as long as the length of the string
					left -= insertBlanks(curBlanks, freesp);
				}
				else{
				//if the string will not fit on oneline then fold it	
					if(toklen > MAXLENGTH){
						folding(state, 1, curLine, left, pos);
						break;
					}
					
					else if(left < MAXLENGTH - leadBlanks){
						freesp = left - toklen; 	
						if(freesp < 1){
							fputc('\n', OUTF);
							left = MAXLENGTH;
							freesp = 0;
						}
						else{
							if(OPTIONS[2]){
								freesp = 1;
							}
							left -= insertBlanks(curBlanks, freesp);
						}
					}
				}
				outputToken(curLine, pos, pos + toklen);
				left -= toklen;
				break;
			}	
		///////////TAGS////////////////
			case('_'):
			{
				state = 5;
				toklen = findTokLen(curLine, pos, 0);
				if( (OPTIONS[7] == 1) || (!formatline) ){
					freesp = strlen(curLine);
					left -= insertBlanks(curBlanks, freesp);
				}
				else{
					if(left < MAXLENGTH - leadBlanks){
						//the tag should be on a new line
						if(!OPTIONS[2]){
							fputc('\n', OUTF);
							left = MAXLENGTH;
						}
					//terse is selected	
						else{
							freesp = left - toklen;
					    //if the token will not fit on the space left or terse is not selected put a new line	
							if(freesp < 1){
								fputc('\n', OUTF);
								left = MAXLENGTH;
								freesp = 0;
							}
							else{ 
								fputc(' ', OUTF);
								left--;
							}
						}
					}
				}
				outputToken(curLine, pos, pos + toklen);
				left -= toklen; 
				break;
			}
			case('l'):
			{
				if (OPTIONS[7] == 1 || (OPTIONS[14] == 1 && strlen(curLine) < MAXLENGTH+1)){
					state = 6;
					toklen = findTokLen(curLine, pos, 0);
					freesp  = strlen(curLine);
					left -= insertBlanks(curBlanks, freesp);
					outputToken(curLine, pos, pos + toklen);
					left -= toklen;
					break;
				}
				if( (!strncasecmp((curLine +pos), "loop_",  5)) && ((!curLine[pos + 5]) || (isspace(curLine[pos + 5]))) && (!OPTIONS[2]) ){
					state = 6;
					NLOOP ++;
					formatLoop(curLine, pos, left, formatline);
					break;
				}
				//else just treat the loop tag as an ordinary data field and go to default
			}
		}//switch
		//reserved words and nondelimited data	
			if (state == 0)
			{
				toklen = findTokLen(curLine, pos, 0);
			    //if format only comments is set then print all of the white spaces
				if( (OPTIONS[7] == 1) || (!formatline) ) {
					freesp = strlen(curLine);
					left -= insertBlanks(curBlanks, freesp);
				}
				else{
					if(toklen > MAXLENGTH){
					    folding(state, 1, curLine, left, pos);
					    break;
					}
					else if(left < MAXLENGTH - leadBlanks){
						freesp = left - toklen;
					//if no space for a blank or the token is a reserved word and terse is not selected(OPTIONS[2] - terse)	insert a new line
						if( (freesp < 1) || ((isreserved) && (!OPTIONS[2])) ){
							fputc('\n', OUTF);
							left = MAXLENGTH;
						}
						else{
						    //terse is selected make freesp equal to 1	
							if(OPTIONS[2]){
								freesp = 1;
							}
							left -= insertBlanks(curBlanks, freesp);
						}		
					}
				}
				outputToken(curLine, pos, pos + toklen);
				left -= toklen;
			}
		state = 0;
		curBlanks = rmfblanks(curLine, pos);
		pos += curBlanks;
		leadBlanks = 0;		
	}
	if(!MAPON){
		if((!OPTIONS[2]) || (!formatline)){
		    if((left != MAXLENGTH) || (!curLine[0])){
			fputc('\n', OUTF);
			left = MAXLENGTH;
		    }
		}
	}
	//if there was nothing on the line, it is not terse and no map is used output a new line
//	else if((!curLine[0]) && (!MAPON) && (!OPTIONS[2]) ){
  //                 fputc('\n', OUTF);
   //                left = MAXLENGTH;
//	}
 if(DEBUG){
 	 fputs("-ENDFORMATLINE-", OUTF);
 }
 
	return 0;	
}
int formatLoop(char * curLine, int &pos, int &left, bool &formatline)
{
if(DEBUG){
  	 fputs("-STARTformatLoop-", OUTF);
}
 
	int  toklen = 0;					//the length of the token	
	bool preserve = false;					//preserve the formating of the loop if the default formating cannot be done and lpTerse is false
	bool lpTerse = (R.getLoop(NLOOP)) ->format; 		//if the loop should be formated tersly		
	int nTags = (R.getLoop(NLOOP)) ->ntags;  		//the number of tags in the loop
	int nData = (R.getLoop(NLOOP)) -> ndata;
	int nItems = nTags + nData;				//the number of items(tags + data) in the loop
	int valid = 0;       					//is the attemted(default) formating valid
	int curCol = -1;					//the current column countig from zero
	int curBlanks = 0;					//the blanks between tokens
	int brk = 0;						//a loop controller
	int state = 0;						//what kind of token is the current
	int adjust = 0;						//the position from where the column strats
	int freesp  = 0;					//the freespace on the output line
	int isFolding = OPTIONS[1];				//indicates whether the loop should be folded or unfolded

	if(nTags != 0){
		valid = (R.getLoop(NLOOP)) -> colsize[nTags];
	}
	//output the loop word-delimiter 
	outputToken(curLine, pos, pos + 5);
	left -= 5;
	//if mapfile is used
	if(MAPON){
		LMCOUNT++;
	}
	//if no map file is used then check whether there is something after the loop if not output a new line
	else{	
		curBlanks = rmfblanks(curLine, pos);
		if(!curLine[curBlanks + pos]){		//if there is something
			fputc('\n', OUTF);
			left = MAXLENGTH;
			fgets(curLine, BUFSIZE, INF);
			LINECOUNT++;
			pos = 0;
			formatline = formatLine();
		}
	}

	do{
	//	curBlanks = rmfblanks(curLine, pos);
	//	pos += curBlanks;
		if((isFolding) && (EXTRAHASHMARK)){
			if(curLine[pos] == '#'){
				fputs("#\n", OUTF);
				left = MAXLENGTH;
			}
			EXTRAHASHMARK = false;
		}
		//strip off the new line character if the token is not a textfield
		if( (curLine[strlen(curLine) -1] == '\n') && (curLine[0] != ';') ){
			curLine[strlen(curLine) -1] = '\0';
			
			//if there is nothin on the line the loop should not be formated tersly and MAPON is not ON and the loop format computed is not valid 
			//so attemt to adhere to the original formatin and output the new line
			if( (!curLine[pos]) && (!MAPON) && (!lpTerse) && (!OPTIONS[2])&& (!valid) ){
				fputc('\n', OUTF);
				left = MAXLENGTH;
			    continue;
			}
		}
		curBlanks = rmfblanks(curLine, pos);
		pos += curBlanks;	 
		if(!formatline){
			//if this is the first item on line and there is something on the output line then put a new line
			if((pos == curBlanks) && (left < MAXLENGTH)){
				fputc('\n', OUTF);
				left = MAXLENGTH;
			}
		}
		//if there is something on the line and curBlanks is 0 we have to make it 1 to insert a space between the last token on the line and the new token to be inserted
		 if ((curBlanks < 1) && (left < MAXLENGTH)&&(!MAPON) ){
			  curBlanks = 1;
		 }
		 
	 	while(curLine[pos]){
			//if mapon is true then the formating is according to a map. Output (if the line and the token should be formated) or skip the neccessary blanks
			if(MAPON){
				if(!formatline){
					insertMapBlanks(left, 0);	//0- do not insert but just skip the blanks
				}
				else{
					insertMapBlanks(left, 1);	//1- insert the blanks according to the map in LINEMAP
				}
			}
			switch (curLine[pos]){
				case(';'):
				{
					//if the semicolumn is at position 0 then it is a text field
					if(pos == 0){
					    nItems --;
					    curCol++;
					    state = 1;
					    if(formatline){	
						FRTEXTCNT++;
						if(isFolding){
 							if(!FORMATTEXT[FRTEXTCNT]){
								outputTextField(curLine, pos, left);
							}
							else{
								folding(state, 0, curLine, left, pos);
							}
						}
					//we are unfolding	
						else{
						//if the textfield is folded and format only comments is not selected unfold it	
							if( /*(!curLine[2])&&*/(curLine[1] == '\\') && (!OPTIONS[7]) ){
								unfoldTextField(curLine, pos, left);
							}
							else{
								outputTextField(curLine, pos, left);
							}
						}
//printf("%d, %d", left, MAXLENGTH);
						if( (left < MAXLENGTH) && (!MAPON) ){
							fputc('\n', OUTF);
							left = MAXLENGTH;
						}
					    }
					    else{
					    	outputTextField(curLine, pos, left);
						if(!MAPON){
							fputc('\n', OUTF);
							left = MAXLENGTH;
						}
					    }
					    formatline = formatLine();
					}
					else{
						state = 0;
					}
					break;
				}
				case('#'):
				{
					state = 2;
					//if foramt everything except comments just ouput the comment 
					if(!formatline) {
						freesp = strlen(curLine);
						left -= insertBlanks(curBlanks, freesp);
						fputs((curLine + pos), OUTF);
						left -= strlen(curLine + pos);
						if(!formatline){
							fputc('\n', OUTF);
							left = MAXLENGTH;
						}
						pos = strlen(curLine);
						break;
					}
					if((pos <= curBlanks)&& (left != MAXLENGTH)){
						fputc('\n',OUTF);
						left = MAXLENGTH;
					}
					toklen = strlen(curLine + pos);
					if(isFolding){
						//is something on the line
						if(left < MAXLENGTH){
							freesp  = left - toklen;
							if((freesp < 1) && (left < 3)){
								fputc('\n', OUTF);
								left = MAXLENGTH;
								freesp = 0;
							}
							else if((freesp < 1) || (lpTerse)){
								freesp = 1;
								curBlanks = 1;
							}
							left -= insertBlanks(curBlanks, freesp);
						}
						if(toklen > left){
							if(curLine[pos + toklen -1] == '\\'){
								strcat(curLine, "\\");	//add an extra hashmark at the end so when unfolding we preserve the comment`s backslash
								EXTRAHASHMARK = true;
							}
								folding(state, 1, curLine, left, pos);
						}
						else{
							outputToken(curLine, pos, pos + toklen);
							left -= toklen;
						}
					}
				//if unfolding 	
					else{
					//	if(toklen < MAXLENGTH){
						     if(!MAPON){	
							freesp = left - toklen;
							if((freesp < 0) || ((freesp == 0) && (left < MAXLENGTH)) ){
								fputc('\n', OUTF);
								left = MAXLENGTH;
								freesp = 0;
							}
							else if((lpTerse) && (left < MAXLENGTH)){
								freesp = curBlanks = 1;
							}
							left -= insertBlanks(curBlanks, freesp);
						     }//if(!MAPON
						     //if the comment is folded unfold it
						     if(curLine[pos +1] == '\\') {
							   //if format everything except comments is not set  
							     if(!OPTIONS[8]){
								unfoldComment(curLine, pos, left);
							     }
							     //if it is set then just ouput the comment do not unfold it
							     else{
							     	outputFoldedComment(curLine, pos, left);
							     }
							     formatline = formatLine();
						     }
						     else{
								outputToken(curLine, pos, pos + toklen);
								left -= toklen;
						     }
					//	}
						//the comment is longer than the maximum allowed length
					//	else{
					//		folding(2, 1, curLine, pos, left);
					//	}
					}
					if( (left < MAXLENGTH) && (!MAPON) ){
						fputc('\n', OUTF);
						left = MAXLENGTH;
					}	
				    break;
				}
				case('\''):
				{
					state = 3;
				//go to the next case since the quoted strings are handled in the same way	
				}
				case('\"'):
				{
					curCol++;
					if( (nItems == nData) && (left < MAXLENGTH) && (formatline) && (!MAPON) ){
						fputc('\n', OUTF);
						left = MAXLENGTH;
					}
					nItems --;
					if(state != 3){
						state = 4;
					}
					toklen = findTokLen(curLine, pos +1, state);
					toklen ++;	//one more for the opening delimiter
				     if(formatline){
					//if the token is longer than the maximum allowed length fold it
					if(toklen > MAXLENGTH) {
						folding(state, 1, curLine, left, pos);
						break;
					}
					else{
					    if(!MAPON){	
					     //if the formating is valid terse is not selected and preserve is not selected
						if((valid) && (!lpTerse) && (!preserve)){
							adjust = (R.getLoop(NLOOP)) -> colsize[curCol];
							//if we are past the correct column position insert a new line
						       	if(MAXLENGTH - left > adjust){
								fputc('\n', OUTF);
								left = MAXLENGTH;
							}
							else{
								adjust -= (MAXLENGTH - left);
							}
							left -= insertBlanks(adjust, adjust);
						}
						else{
							if(left < MAXLENGTH){
								freesp = left - toklen;
								if(freesp < 1){
									fputc('\n', OUTF);
									left = MAXLENGTH;
									freesp = 0;
								}
								else{
									//if the loop should be formated tersly and freesp is bigger than one reduce it to one
									if(lpTerse){
										freesp = 1;
										curBlanks = 1;
									}
									left -= insertBlanks(curBlanks, freesp);	
								}
							}
						}
					    }//if(!MAPON)
					}
				     }//if(formatline)
				     else{
					     left -= insertBlanks(curBlanks, curBlanks);
				     }
				     outputToken(curLine, pos, pos + toklen);
				     left -= toklen;
				   break;
				}
			////TAGS////////	
				case('_'):
				{    	
					nItems --;
					state = 5;
					if( (formatline) && (!OPTIONS[7]) ){
						if( (left < MAXLENGTH) && (!MAPON) ){
							fputc('\n', OUTF);
							left = MAXLENGTH;
						}
					}
					else{
						left -= insertBlanks(curBlanks, curBlanks);
					}
					toklen = findTokLen(curLine, pos, 0);
					outputToken(curLine, pos, pos + toklen);
					left -= toklen;
				     break;
				}
			}//switch	
				if(state == 0)
				{
					curCol++;
					if( (nItems == nData) && (left < MAXLENGTH) && (formatline) && (!MAPON)&&(!OPTIONS[7]) ){
						fputc('\n', OUTF);
						left = MAXLENGTH;
					}
					nItems --;
					state = 0;
					toklen = findTokLen(curLine, pos, 0);
				    if(formatline){	
					if(toklen >MAXLENGTH){
						folding(state, 1, curLine, left, pos);
						break;
					}
					else{
					      if(!MAPON){	
						//if the loop formating is valid compute the number of blanks to output to align the item in its column
						if((valid) && (!lpTerse) && (!preserve)){
							adjust = (R.getLoop(NLOOP)) -> colsize[curCol];
							if(MAXLENGTH - left > adjust){
								fputc('\n', OUTF);
								left = MAXLENGTH;
							}
							else{
								adjust -= (MAXLENGTH - left);
							}
							left -=	insertBlanks(adjust, adjust);	
						}
						else{
						     //if format only comments is not set	
							if(left < MAXLENGTH){
								freesp = left - toklen;
								if(freesp < 1){
									fputc('\n', OUTF);
									left = MAXLENGTH;
									freesp = 0;
								}
								else{
								 
									//if the loop should be formated tersly and freesp is bigger than one reduce it to one	
									if(lpTerse){
										freesp = 1;
										curBlanks = 1;
									}
									left -= insertBlanks(curBlanks, freesp);
								}
							}
						}
					     }//if(!MAPON)	
					}
				    }//if formatline
				    else{
				    	left -= insertBlanks(curBlanks, curBlanks);
				    }
				    outputToken(curLine, pos, pos + toklen);
				    left -= toklen;
				}
		if(nItems < 0){
			brk = 1;
			break;
		}
		if(MAPON){
			LMCOUNT++;
			if(!LINEMAP[LMCOUNT]){
				fgets(LINEMAP, BUFSIZE, MAPFILE);
				LMCOUNT = 0;
			}
			if( (LINEMAP[LMCOUNT] == '\'') || (LINEMAP[LMCOUNT] == '\"') || (LINEMAP[LMCOUNT] == ';') ){
				 LMCOUNT++;
			}
		}			
		//if curCol is equal to the last column then the next column should be column zero 
		if(curCol == nTags -1){
			if((left < MAXLENGTH) && (!lpTerse) && (formatline) && (!MAPON) ){	//check if there is something on the line output new line
				fputc('\n', OUTF);
				left = MAXLENGTH;
			}
			curCol = -1;				//-1 since curCol is first incremented when a token is found 
		}
		state = 0;		//default state
		//Compute the blanks between the token
		curBlanks = rmfblanks(curLine, pos);
		pos += curBlanks;		
	   } //while (curLine[pos])
	//if we have ouputed all items from the loop	
     	  if(brk){
		  break;
	  }
	
	  //if the loop shoud not be formated tersely, the formating is not valid, the length of the line is less than the maximum allowed length and there is something on the line
  	  if(((!lpTerse) && (!valid) && (strlen(curLine) <= MAXLENGTH) && (left < MAXLENGTH)) || (!formatline)){
	     if(!MAPON){	  
		fputc('\n', OUTF);
		left = MAXLENGTH;
	     }
	  }		
	  pos = 0;
	  LINECOUNT++;
	  if(MAPON){
		//if the previous line was formated and the new line should not be then make sure it starts on a new line
		if( (formatline) && (!formatLine()) ){
			if(left<MAXLENGTH){
				fputc('\n', OUTF);
				left = MAXLENGTH;
			}
		}
	  } 
	  formatline = formatLine();
	  if( (!formatline) && (left < MAXLENGTH) ){
	  	  fputc('\n', OUTF);
		  left = MAXLENGTH;
	  }
	}while(fgets(curLine, BUFSIZE, INF));
	if(DEBUG){
		fputs("-ENDformatLoop-", OUTF);
	}
    return 0;
}
int unfoldline(char *curLine, int &pos, int &left)
{
 if(DEBUG){ fputs("-unfoldline-", OUTF); }
 
	int leadBlanks = 0;	//the number of leading blanks inserted
	int lblanks = 0;	//the number of leading blanks on the line
	int curBlanks = 0;	//blanks between tokens
	int freesp = 0;		//the freespace on the line
	int toklen = 0;		//the length of the token
	int state = 0;		//indicates the type of token 0-nondelimted, 1-textfield, 2-comment, 3 and 4 delimeted (' or ") 	
	bool formatline = formatLine();
	bool newlineignore = false;	//if the new line should be ignored
	bool newlineinsert = false;	//if an additional new line should be inserted
       	
	if(pos == 0){
		    if( (!formatline) && (left < MAXLENGTH) ){
		    	fputc('\n', OUTF);
			left = MAXLENGTH;	
		    }
		    lblanks = rmfblanks(curLine, pos);
		    pos = lblanks;
		    if(curLine[0] != ';'){
			    rmtblanks(curLine);
		    }
		    else rmttblanks(curLine);	
		if( ((OPTIONS[4] == 1)&&(!MAPON)) || (!formatline) ) {
			leadBlanks = insertBlanks(lblanks, left);
			left -= leadBlanks;
		}
		else{
			leadBlanks = 0;
		}
	}
	else{
		curBlanks = rmfblanks(curLine, pos);
		rmtblanks(curLine);
	}
	//if there is something on the line and terse is selected there should be at least one blank character outputed
	if((left < MAXLENGTH) &&(!MAPON)){
		curBlanks = 1;
	}
	//if there is nothing on the line map is not on and terse is not selected output a new line 
	if((!curLine[pos]) && (!MAPON) && (!OPTIONS[2]) ){
		//if there is something on the line then put a new line before inserting the current new line
		if(left < MAXLENGTH){
			fputc('\n', OUTF);
		}
		fputc('\n', OUTF);
		left = MAXLENGTH;
	}
	
	while(curLine[pos]){
	//insert the proper number of blanks if MAPON is true	
		if(MAPON){
			//if the line should not be formated
			if(!formatline){	
				insertMapBlanks(left, 0);
			}
			else{
				insertMapBlanks(left, 1);
			}
		}
		state = 0;
		switch(curLine[pos])
		{
			case(';'):
			{
				if(pos == 0){
					
					state = 1;
					//if the textfield is folded and format only comments is not selected
					if( (curLine[pos +1] == '\\') && (!OPTIONS[7]) && (formatline) ){
						unfoldTextField(curLine, pos, left);
					}
					else{
						//pass formatline to the method as well also insert the blank characters 
						outputTextField(curLine, pos, left);
					}
					if((left < MAXLENGTH)&&(!MAPON)){
						fputc('\n', OUTF);
						left = MAXLENGTH;
					}
				}
				else{
					state = 0;
				}
			   break;
			}
			case('#'):
			{
				state = 2;
				//if the comment is on a new line and there is something on the current line output a new line
				if( (pos == lblanks) && (left < (MAXLENGTH - lblanks) ) && (!MAPON) && (formatline) ){
					fputc('\n', OUTF);
					curBlanks = 0;
					left = MAXLENGTH;
				}
			    //if the comment is folded unfold it	
				if( (curLine[1] == '\\') && (formatline) ){
				     if(!OPTIONS[8]){	
					//if there is something on the line output 1 empty  character
					if(left < (MAXLENGTH - leadBlanks) && (!MAPON) ) {
						fputc(' ', OUTF);
						left --;
					}
					unfoldComment(curLine, pos, left);
				     }
				     else{
				     	outputFoldedComment(curLine, pos, left);
				     }
				}
				else{
				     toklen = strlen(curLine + pos);
				     if( (formatline) && (!MAPON) ){	
					freesp = left - toklen;
					if( ((freesp == 0) && (left < MAXLENGTH - leadBlanks)) || (freesp < 0)){
						fputc('\n', OUTF);
						left = MAXLENGTH;
						freesp = 0;
					}
					//if terse is selected
					else if(OPTIONS[2] == 1){
						if(left < (MAXLENGTH - leadBlanks)) {
							freesp = 1;
						}
						else{
							freesp = 0;
						}
					}
				     }
				     else{
				     	freesp = curBlanks;
				     }
				     if(!MAPON){
					left -= insertBlanks(freesp, curBlanks);
				     }
				     outputToken(curLine, pos, pos + toklen);
				     left -= toklen;
				}
				// ouput new line otherwise the next token will be put after the comments on the same line and will look like it is part of the comment
				if((left != MAXLENGTH) && (formatline) && (!MAPON)) {
					fputc('\n', OUTF);
					left = MAXLENGTH;
				}
			   break;
			}
			case('\''):
			{
				state = 3;
			  //go to the next case since the quoted strings are handled in the same way	
			}
			case('\"'):
			{
				if(state != 3){
					state = 4;
				}
				toklen = findTokLen(curLine, pos +1, state);
				toklen ++; 	//add one for the opening delimiter
				//if format only comments is set do not format the token
				if(!formatline){
					//if this is the first element on the line and there is something on the output line put a new line
				 	if( ((pos - lblanks) <= 1) && (left < MAXLENGTH - leadBlanks) && (formatline) ){
				       		 fputc('\n', OUTF);
						 left = MAXLENGTH;
						 curBlanks = 0; 
					 }					
				   freesp = strlen(curLine);
				   left -= insertBlanks(curBlanks, freesp);
				}
				else if(!MAPON){
					freesp = left - toklen;
					if(freesp < 1){
					//if this is not the first item on the line and freesp is 0(that means it will fit exactly on the line) output a new line
						if((freesp < 1) || ((freesp ==0) && (left < (MAXLENGTH - leadBlanks))) ){
							fputc('\n' ,OUTF);
							left = MAXLENGTH;
							freesp = 0;
						}
					}
					else{
						//if terse unfolding is set
						if(OPTIONS[2]){
							freesp = 1;
						}
						//if the column with respect to which the data should be aligned is set and the data will fit on the line if 
						//it is aligned according to column OPTIONS[9]
						else if( (OPTIONS[9]) && ((OPTIONS[9] + toklen) <= MAXLENGTH) ){
							curBlanks = OPTIONS[9] - (MAXLENGTH - left);
						//if we are already past that column and there is something on the line 
						//output just one space character so we are as close as possible to the column	
							if((curBlanks < 1) && (left < MAXLENGTH - leadBlanks)){
								curBlanks = 1;
							}	
						}
					}
					left -= insertBlanks(curBlanks, freesp);
				}
				outputToken(curLine, pos, pos + toklen);
				left -= toklen;
			   break;
			}
			case('_'):
			{
				state = 5;
				toklen = findTokLen(curLine, pos, 0);
				//if format only comments is set
				if(!formatline) {
					 //if this is not the first item on the line and freesp is 0(that means it will fit exactly on the line) output a new line
					if( ((pos - lblanks) <= 1) && (left < MAXLENGTH - leadBlanks) && (formatline) ){
						fputc('\n', OUTF);
						left = MAXLENGTH;
					}
					freesp = strlen(curLine);
					left -= insertBlanks(curBlanks, freesp);
				}
				else if(!MAPON){
					//if terse is not set and there is something on the line put a new line
					if (!OPTIONS[2]){
					       if(left < (MAXLENGTH - leadBlanks)){
							fputc('\n', OUTF);
							left = MAXLENGTH;
					       }
					}
					else {
						freesp = left - toklen;
						if( (freesp < 1) && (!((freesp == 0) && (left < (MAXLENGTH - leadBlanks))) ) ){
							fputc('\n', OUTF);
							left = MAXLENGTH;	
						}
						else{
							freesp = 1;
							left -= insertBlanks(curBlanks, freesp);
						}
					}	
				}
				outputToken(curLine, pos, pos + toklen);
				left -= toklen;
			   break;
			}
			case('l'):
			{
				//if this is a loop tag and terse is not set
				if( (!strncasecmp((curLine + pos), "loop_", 5)) && ((!curLine[pos + 5]) || (isspace(curLine[pos + 5]))) && (!OPTIONS[2]) ){
				      state = 6;
				      NLOOP++;
				     	if( (left < (MAXLENGTH - leadBlanks)) && (!OPTIONS[2]) && (!MAPON) && (formatline)){ 
				      		fputc('\n', OUTF);
						left = MAXLENGTH;
				      	}
				     	 //if the line should not be formated ouput the leading blanks if any
				      	if(!formatline){
				       	    left -= insertBlanks(curBlanks, curBlanks);
				      	}
				      formatLoop(curLine, pos, left, formatline);  
				}
			   break;
			}	
		}	//switch curLine[pos]
		//if state is still 0 then it is nondelimeted datafield
		if(state == 0){
			toklen = findTokLen(curLine, pos, 0);
				freesp = left - toklen;
				//if is reserved and not terse and there is something on the line output a new line
				if(isreserved){
					if((!OPTIONS[2]) && (left < (MAXLENGTH - leadBlanks)) && (formatline) && (!MAPON) ) {
						putc('\n', OUTF);
						left = MAXLENGTH;
					}
				}
				else if( (!MAPON) || (!formatline) ){
				     if(formatline){			
					if((freesp < 1) && (!((freesp == 0) && (left >= (MAXLENGTH - leadBlanks))) ) ){  
						fputc('\n', OUTF);
						left = MAXLENGTH;
						freesp = 0;
					}
					else if(OPTIONS[2]){
						freesp = 1;
						if(left < (MAXLENGTH - leadBlanks)){
							curBlanks =1;
						}
					}
					//terse is not set
					else{
						if( (OPTIONS[9]) && ((OPTIONS[9] + toklen) <= MAXLENGTH) ){
							curBlanks  = OPTIONS[9] - (MAXLENGTH - left);
							if( (curBlanks < 1) && (left <( MAXLENGTH - leadBlanks))){
								curBlanks = 1;
							}
						}
					}
				     }
				     else{
				     	freesp = curBlanks;
				     }
				     left -= insertBlanks(curBlanks, freesp);
				}
			outputToken(curLine, pos, pos + toklen);
			left -= toklen;
		}
		curBlanks = rmfblanks(curLine, pos);
		pos += curBlanks;
		leadBlanks = 0;
		if(MAPON){
			LMCOUNT++;
			if( (LINEMAP[LMCOUNT] == ';') || (LINEMAP[LMCOUNT] == '\'') || (LINEMAP[LMCOUNT] == '\"')){
				LMCOUNT++;
			}
		}
	}//while(curLine[pos])
//	if(!MAPON){
//		if( (left < MAXLENGTH) && (!curLine[pos])){
//		}	
//	}
	if(DEBUG){fputs("-outunfoldLine-", OUTF); }
	return 0;
	
}
void unfoldComment(char *curLine, int &pos, int &left){
	 if(DEBUG){ fputs("-unfoldComment-", OUTF);}
	 
	 char *buffer = new char[BUFSIZE];	 
	 int len = 0;
	 int bufflen = 0;
	 bool endComment = false;

	 //put the opening delimiter in the buffer
	  strcpy(buffer, "#");
	  LINECOUNT++;
 	  fgets(curLine, BUFSIZE, INF);
	  do{
		  rmtblanks(curLine);
		  if(curLine[0] == '#'){
		  	if(curLine[strlen(curLine) -1] == '\\'){
		  		curLine[strlen(curLine) -1] = '\0';
		  	}
			else{
				endComment = true;
				pos = strlen(curLine);
			}
		  	strncat(buffer, (curLine +1), (BUFSIZE - strlen(curLine)));
		  }
		  else{
		  	endComment = true;
			pos = 0;
		  }
		  if(endComment){
		  	rmtblanks(buffer);
		 	bufflen = strlen(buffer);
		 	//check for valididty if the string is longer than MAXLENGTH fold it back
		  	if(bufflen > MAXLENGTH){
		  		folding(2, 1, buffer, left, pos);
		  	}
		  	else{
		  		if(bufflen > left){
					fputc('\n', OUTF);
				}
				fputs(buffer, OUTF);
		  	}
			if(!MAPON){
				fputc('\n', OUTF);
				left = MAXLENGTH;
			}
		     break;	
		  }
	      LINECOUNT++;
	  }while(fgets(curLine, BUFSIZE, INF));	  
	 if(DEBUG){ fputs("-outunfoldComment-", OUTF); }
	  
	  delete [] buffer;
}
int unfoldTextField(char *curLine, int &pos,int &left)
{	
    vector <char> TextVector;
	if(DEBUG){ fputs("-unfoldTextField-", OUTF);}
	bool newLine = false;
	bool converttoString = false;
	bool mustrefold = false;
	bool mustbetext = false;
	bool startwithnl = false;
	int oleft;
	char delimiter = ';';
	bool formatFirstLine = formatLine(); //indicates wether a new line should be inserted at the end of the textfield
					//used when unfolding MAPON is true and format only chucks of the file is true then if the file chunk ends in the middle of the comment we should output a newline 
	int blanks = 0;
	oleft = left;
	
	TextVector.clear();
	
	delimiter = ';';
	if(MAPON){
		if( (LINEMAP[LMCOUNT +1] == '\'') || (LINEMAP[LMCOUNT +1] == '"') || (LINEMAP[LMCOUNT +1] == ';') ){
			delimiter = LINEMAP[LMCOUNT +1];
		}
		else{
			delimiter = ' ';
		}	
	}
	
	mustrefold = false;
	mustbetext = false;
	startwithnl = false;
	
	//output the opening delimiter if the textfield should not be converted to a string(only if MAPON is true)
	//If the file should be mapped then check whether the current position is at the begining of a line 
	//  else the text field should be converted to a string	
	if( (delimiter == ';') && (left < MAXLENGTH)){
		/* fputc('\n', OUTF); */
		startwithnl = true;
		left = oleft = MAXLENGTH;
	}
	//output the opening delimiter
	if(delimiter != ' '){
		/* fputc(delimiter, OUTF); */
		TextVector.push_back(delimiter);
		left --;
	}
	//the delimiter has been outputed read the next line
	fgets(curLine, MAXLENGTH, INF);
	LINECOUNT++;
	do{
		//get rid of the new line character
		if(curLine[strlen(curLine) -1] == '\n'){
			curLine[strlen(curLine) -1] = '\0';
		}
		//if the closing delimiter is found output it and return
		if( (curLine[0] == ';') && ((isspace(curLine[1])) || (!curLine[1])) ){
			/* fputc(delimiter, OUTF);	*/ //output the closing deliemter
			TextVector.push_back(delimiter);
			left = MAXLENGTH -1;
			pos = 1;
			//if MPON and the file chunck that is being formated ended in the textfield then a new line should be outputed at the end 
			if( (MAPON)&&((formatFirstLine) && (!formatLine())) ){
					/* fputs((curLine +1), OUTF); */
					{ int ick=1;
					  char icc;
					  while((icc=*(curLine+ick++))){
					    TextVector.push_back(icc);
					  }
					}
					left -= strlen((curLine +1));
					pos += strlen((curLine+1));
			}
			else{
				rmtblanks(curLine);
			}
		   break;
		}
//		rmtblanks(curLine);
		//if the character before the newline character is a backslash then unfold
		if(curLine[strlen(curLine) -1] == '\\'){
			curLine[strlen(curLine) -1] = '\0';
		}
		//if the textfield should not be converted to a string. 
		//if it should and there is a new line then it is an error
		//which we recover from by leaving it as a text field
		else /* if(delimiter == ';') */{
		    mustbetext = true;	
			strcat(curLine, "\n");
			newLine = true;
		}
		/* fputs(curLine, OUTF); */
		{ int ick=0;
		  char icc;
		  while((icc=*(curLine+ick++))){
		    if (icc == '\n') 
		    {  mustbetext = true;
		       left = MAXLENGTH;
		    }
			TextVector.push_back(icc);
			left--;
			if (left < 0) 
			{
			    if (!mustbetext && oleft < MAXLENGTH)
			    {
			      left += (MAXLENGTH-oleft);
			      oleft = MAXLENGTH;
			      startwithnl = true;
			    }
			    else 
			    {
				  mustrefold = true;
				  mustbetext = true;
				  left = MAXLENGTH-1;
			    }
			}
		  }
		}
		/* left -= strlen(curLine); */
//Some checks with left may be made here to see if the outputed line has valid length.
		if(newLine){
			left = MAXLENGTH;
			newLine = false;
		}
	   LINECOUNT++;
	} while(fgets(curLine, BUFSIZE, INF));
	
	if (delimiter == ';' && !mustbetext )
	{  int ick;
	   char icc;
	   bool nodq, nosq;
	   nodq = false;
	   nosq = false;
	   for(ick=0; (ick<TextVector.size()) && (!nosq || !nodq); ick++) 
	   {
	     if (!nosq && TextVector.at(ick)=='\'')
	     {
	     	if (ick < TextVector.size()-1) 
	     	{
	     	  if (isspace (TextVector.at(ick+1))) nosq = true;
	     	}
	     }
	     if (!nodq && TextVector.at(ick)=='"')
	     {
	     	if (ick < TextVector.size()-1) 
	     	{
	     	  if (isspace (TextVector.at(ick+1))) nodq = true;
	     	}
	     }
	   }  
	   if (nosq && nodq) mustbetext = true;
	   if (!mustbetext) 
	   {
	     delimiter = '\'';
	     left = oleft;
	     if (nodq) delimiter = '"';
	     if (startwithnl) 
	     {
	       left = MAXLENGTH;
	       fputc('\n',OUTF); 	
	     }
	     for(ick=0; ick<TextVector.size(); ick++) 
	     {
	      icc=TextVector.at(ick);
	      if (ick ==0 && icc==';') icc= delimiter;
	      if (ick == TextVector.size()-1) icc = delimiter;
	      fputc(icc,OUTF);
	     }
	     if (left > 0) 
	     {
	     	fputc(' ',OUTF);
	     	left--;
	     }
	     if(DEBUG){ fputs("-outunfoldTextField-", OUTF); }
         return 0;
	   }
	}
	
	if (mustrefold) 
	{  int ick;
	   char icc;
	   if (oleft < MAXLENGTH || startwithnl) 
	   {  fputc('\n',OUTF);
		  left = MAXLENGTH;
	   }
	   for (ick=0; ick<TextVector.size()-1; ick++) 
	   {
	   	  icc=TextVector.at(ick);
	   	  if (ick ==0 ) 
	   	  {
	   	  	 fputs(";\\\n",OUTF);
	   	  	 if (delimiter == ' ') 
	   	  	 {
	   	  	 	fputc(icc, OUTF);
	   	  	 	left = MAXLENGTH-1;
	   	  	 }
	   	  }
	   	  else 
	   	  {
	   	  	 fputc(icc,OUTF);
	   	  	 left--;
	   	  	 if (left < 1) 
	   	  	 {
	   	  	 	fputs("\\\n", OUTF);
	   	  	 	left = MAXLENGTH;
	   	  	 }
	   	  }
	   }
	   fputs("\n; ", OUTF);
	   left = MAXLENGTH-2;
       if(DEBUG){ fputs("-outunfoldTextField-", OUTF); }
       return 0;
	}
	{  int ick;
	   char icc;
	   if (startwithnl) fputc('\n', OUTF);
	   for(ick=0; ick<TextVector.size(); ick++) fputc((icc=TextVector.at(ick)),OUTF);
	}
	if(DEBUG){ fputs("-outunfoldTextField-", OUTF); }
   return 0;
}

int outputToken(char *curLine, int &pos, int end)
{
	if(DEBUG){
		fputs("-STARToutputToken-", OUTF);
	}
	if(pos < end){
		for(pos; pos < end;pos++)
		{
			fputc(curLine[pos], OUTF);
		}
	}
	if(DEBUG){
 		fputs("-ENDoutputToken-", OUTF);
	}
	return 0;
}
int folding(char state, char once, char *curLine, int &left, int &pos)
{
	if(DEBUG){
		fputs("-STARTfolding-", OUTF);
	}
	
//Implements the folding protocol of long tokens 
	int brk = 0;
	bool formatText = false;
//	char **tokens;

    /* fprintf(stderr,"state %d, once %d, curLine %s, left %d, pos %d\n", state, once, curLine, left, pos); */
	//if a text field output
	if(state == 1){
		if(left < MAXLENGTH){
		     fputc('\n', OUTF);
		}
		fputs(";\\\n", OUTF);
		left = MAXLENGTH;
		pos++;
		do
		{
		
			if((pos == 0) && (curLine[pos] == ';') && (isspace(curLine[pos+1]) || !curLine[pos+1]))
			{
				rmtblanks(curLine);		//remove the trailing blanks new added february 07
				fputc(';', OUTF);
				pos++;
				left --;
				break;
				
			}
			else{  
				int savbpos, savbleft, savipos;
				char savc;

				savipos = pos;
				savbpos = pos-1;
				savbleft = left+1;
				for(pos; curLine[pos]; pos++)
				{
					if( curLine[pos] == ' ' || curLine[pos] == '\t' ) {
					  savbpos = pos;
					  savbleft = left;
					}
				//check for ; at the position where the line will be folded
				//if it is ; it may cause ambiguity with the closing delimiter of the text 
				//so fold the line one character earlier 	
					if(((left == 2) && (curLine[pos +1] != '\n') && (curLine[pos +2] != '\n'))
					   || (left == MAXLENGTH-1 && curLine[pos-1]==';' && isspace(curLine[pos]))){
						if(savbpos < pos && savbpos > savipos) {
							pos = savbpos;
							left = savbleft;
						}
						if (savipos < pos) {
							savc = curLine[pos];
							curLine[pos] = '\0';
							fputs(curLine+savipos,OUTF);
							curLine[pos] = savc;
							savipos = pos;
						}
						if(curLine[pos +1] != ';'){
							fputc(curLine[pos], OUTF);
							savipos++;
							pos++;
							left--;
						}
						fputs("\\\n", OUTF);
				        left = MAXLENGTH;
				        savbpos = pos-1;
				        savipos = pos;
				        savbleft = left;
					}
					if(curLine[pos] == '\n'){
						savc = curLine[pos];
						curLine[pos] = '\0';
						fputs(curLine+savipos,OUTF);
						curLine[pos] = savc;
						savbpos = pos-1;
						savipos = pos;
						savbleft = left;
						//check for a backslash at the end of line and output another one and a newline if found
						if(curLine[pos -1] == '\\'){
							if(left <= 0){
								fputs("\n\\", OUTF);
							}
							fputs("\\\n", OUTF);
						   	left = MAXLENGTH;	
						}
					}
					// fputc(curLine[pos], OUTF);
					left--;
				}
				if (savipos < pos) {
					fputs(curLine+savipos,OUTF);
					savipos = pos;
				}
			}
		  pos = 0;
		  left = MAXLENGTH;
		  LINECOUNT++;
		}while(fgets(curLine, BUFSIZE, INF));
	}
    //if a comment
	else if(state == 2){
		int savbpos, savbleft, savipos;
		char savc;

		rmtblanks(curLine); 
		if(left < 2){
			fputc('\n', OUTF);
			left = MAXLENGTH;
		}
		fputs("#\\\n#", OUTF);
		left = MAXLENGTH -1;
		pos++;
		savipos = pos;
		savbpos = pos-1;
		savbleft = left+1;

		for(pos; curLine[pos];pos++)
		{
			if(curLine[pos] == ' ' || curLine[pos] == '\t' ) {
			savbpos = pos;
			savbleft = left;
			}

			if(left == 1){
			    if (savbpos < pos && savbpos > savipos) {
				  pos = savbpos+1;
				  left = savbleft-1;
				}
				savc = curLine[pos];
				curLine[pos] = '\0';
				fputs(curLine+savipos,OUTF);
				curLine[pos] = savc;
				savipos = pos;
				fputs("\\\n#", OUTF);
				left = MAXLENGTH -1;
			}
			left--;
		}
		if (savipos < pos) {
			fputs(curLine+savipos,OUTF);
			savipos = pos;
		}
	}	
	else if( (state == 3) || (state == 4) ) {
		int savbpos, savbleft, savipos;
		char savc;

		char delimiter = '\'';
		if(state == 4){
			delimiter = '\"';
		}
		pos++;		//skip the opening delimiter
		//output the opening delimiter
		if(left < MAXLENGTH){
			fputc('\n', OUTF);
		}
		fputs(";\\\n", OUTF);
		left = MAXLENGTH;

		savipos = pos;
		savbpos = pos-1;
		savbleft = left+1;
		for(pos; curLine[pos]; pos++)
		{
			if( curLine[pos] == ' ' || curLine[pos] == '\t' ) {
				savbpos = pos;
				savbleft = left;
			}
			if((curLine[pos] == delimiter) && ((!curLine[pos + 1]) || (isspace(curLine[pos + 1]))) ){
			    if (savipos < pos) {
					savc = curLine[pos];
					curLine[pos] = '\0';
					fputs(curLine+savipos,OUTF);
					curLine[pos] = savc;
					savipos = pos;
				}
				//if the last character was backslash then output another one and a new line
				if(curLine[pos -1] == '\\'){
					if(left <= 0){
						fputs("\n\\", OUTF);
					}
					fputs("\\\n", OUTF);
				}
				fputs("\\\n;\n", OUTF);
				left = MAXLENGTH;
				pos++;
				break;
				
			}
			if(((left == 2) && (curLine[pos +1] != delimiter) && (curLine[pos +2] != delimiter))
			  || (left == MAXLENGTH-1 && curLine[pos-1]==';' && isspace(curLine[pos]))){
				if(savbpos < pos && savbpos > savipos) {
					pos = savbpos;
					left = savbleft;
				}
				if (savipos < pos) {
					savc = curLine[pos];
					curLine[pos] = '\0';
					fputs(curLine+savipos,OUTF);
					curLine[pos] = savc;
					savipos = pos;
				}
				if(curLine[pos +1] != ';'){
					fputc(curLine[pos], OUTF);
					savipos++;
					pos++;
					left--;
				}
				fputs("\\\n", OUTF);
				left = MAXLENGTH;
				savbpos = pos-1;
				savipos = pos;
				savbleft = left;
			}
			left --;
		}
		if (savipos < pos) {
		    if (curLine[pos-1] == delimiter) {
				if (savipos < pos -1) {
				  curLine[pos-1] = '\0';
				  fputs(curLine+savipos,OUTF);
				  savipos = pos;
				  curLine[pos-1] = delimiter;
				}
			} else {
				fputs(curLine+savipos,OUTF);
				savipos = pos;
			}
		}
	}
	else{
		if(left < MAXLENGTH){
			fputc('\n', OUTF);
			left = MAXLENGTH;
		}
		fputs(";\\\n", OUTF);
		
		for(pos; curLine[pos]; pos++)
		{
			if((left == 1) && (!((curLine[pos + 1]) && (isspace(curLine[pos + 1])))) ) {
				fputs("\\\n", OUTF);
				left = MAXLENGTH;
			}
			fputc(curLine[pos], OUTF);
			left--;
			if( (!curLine[pos +1]) || (isspace(curLine[pos])) ){
				if(curLine[pos -1] == '\\'){
					if(left <= 0){
						fputs("\n\\", OUTF);
					}
					fputs("\\\n", OUTF);
				}
				if(left < MAXLENGTH){
					fputc('\n', OUTF);
					left = MAXLENGTH;
				}
				fputs(";\n", OUTF);
				left = MAXLENGTH;
				break;
			}
		}
	}
	if(DEBUG){
		fputs("-ENDfolding-", OUTF);
	}
	return 0;
}

bool formatLine(){
	//Check whether only blocks of the file should be formated and returns true if the line is in the block false otherwise or if the whole file should be formated
	bool formatline = true;
	
	 if(OPTIONS[6] == 0){
		  if(START == -1){
			formatline = false;
		  }
		  else if(START == 0){
			 //if this was there are not more chunks in the array then the rest of the file should not be formated 
			if(FChIterator >= FileChunks.end()){ 
			       START = -1;
				formatline = false;
			}
			else{
			   START = *FChIterator;
			   FChIterator++;
			   END = *FChIterator;
			   FChIterator++;
			}
		 }
		  
		 if(LINECOUNT >= START){
			if(LINECOUNT >= END){
				       START = 0;
			       END = 0;
			       formatline = false;
			}
		 }
		 else{
			formatline = false;
		 }
	  }

	 return formatline;
}
void outputFoldedComment(char *curLine, int& pos, int &left)
{
	do{
		if(curLine[strlen(curLine) -1] != '\\'){
			if(MAPON){
				rmtblanks(curLine);
				left -= strlen(curLine);
			}
			else{
				left = MAXLENGTH;
			}
			fputs(curLine, OUTF);
			pos  = strlen(curLine);
			
		}
		fputs((curLine + pos), OUTF);
		LINECOUNT++;
	}while(fgets(curLine, BUFSIZE, INF));		
}
void checkNewLineMAP(int &left)
{
	while(LINEMAP[LMCOUNT] == 'n')
	{
		fputc('\n', OUTF);
		LMCOUNT++;
		left = MAXLENGTH;
	}
}
