/*
CIFFOLD 0.5.3 Pre-Release
by Kostadin Mitev, Georgi Todorov and Herbert J. Bernstein

Copyright (C) Kostadin Mitev 2005

Work funded in part by the International Union of Crystallography under
a grant to Dowling College.

This software is covered by the GNU General Public License 
<see COPYING>.

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

 */



#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <iostream>
#include <cstring>
#include "Globals.h"
#include "getOpt.h"

using namespace std;
	
	void printHelp();
	
	int parseOpt (int argc, char **argv, char **pinfile)
	{
	  int ret = 1;	
	  static string version = "ciffold 0.5 Pre-Release";
	  int aflag = 0;
	  int bflag = 0;
	  char *inputfile = new char[255];
	  inputfile[0]='\0';
	  char *outputfile = new char[255];
	  outputfile[0]='\0';
	  char *tmpstr = NULL;
	  int index;
	  int c;
	  MAXLENGTH = 80;	//initialize maxlength

	  
  	  //initialize evertything to -1
	  for(int i = 0; i < 15;i++)
	  {
		  OPTIONS[i] = -1;
	  }	  
	  opterr = 0;
	
	  while ((c = getopt (argc, argv, "cdegwnuthMVLi:x:o:l:m:C:p:v:")) != -1)
	    switch (c)
	    {
	      case ('c'):
	      {
	       	 OPTIONS[7] = 1;	//format only comments
	        break;
	      }
	      case ('d'):
	      {
	        OPTIONS[10] = 1;	//dictionary
	        break;
	      }
	      case ('e'):
	      {
		OPTIONS[8] = 1;    	//format everything except comments
	        break;
	      }
	      case ('g'):
	      {
		ret = 2;		//invoke the GUI menu
	      	break;
	      }
	      case('M'):
	      {
		   OPTIONS[13] = 1;	//generate a MAP
		      break;
	      }
	      case('V'):
	      {
		 cout<<version<<endl;	 //print ciffold`s version
		 exit(0);
	      }
	      case ('w'):
	      {
		      OPTIONS[1] = 1;	//folding
	      	      break;
	      }
	      case ('n'):
	      {
		      OPTIONS[14] = 1;	//minimal folding
		      if (OPTIONS[1] == -1 ) OPTIONS[1] = 1;
		      if (OPTIONS[1] == 0 ) {
			printHelp();
			exit(0);
		      } 
	      	      break;
	      }
	      case ('u'):
	      {
		      OPTIONS[1] = 0;	//unfolding
	      	      break;
	      }
	      case ('L'):
	      {
		      OPTIONS[4] = 1;	//preserve leading blanks
	      	      break;
	      }
	      case ('t'):
	      {
	      	OPTIONS[2] = 1;		//terse formatting
		break;
	      }
	      case('h'):				//print help message and exit
	      {
		printHelp();
		exit(0);
		break;
	      }
	      case ('i'):
	      {
		      strncpy(inputfile, optarg, 255);	//the name of the input file
	      	      break;
	      }
	      case ('x'):
	      {
		      if(tokanize(optarg)){
			      OPTIONS[6] = 0;	//process the intire file (0- do not process the entire file)
		      }
		      else{
		      	OPTIONS[6] = 1;
		      }
		      break;
	      }
	      case ('o'):
	      {
		      strncpy(outputfile, optarg, 255);	//the name of the outputfile
	      	      break;
	      }
	      case ('l'):
	      {
		      OPTIONS[3] = 1;		  //terse folding on loops
		      OPTIONS[5] = atoi(optarg);  //put the loops size here 
		      break;
	      }
	      case('m'):
	      {
		MAXLENGTH = atoi(optarg);		//the maximum line length
	  	break;	
	      }
	      case ('C'):
	      {
		      OPTIONS[9] = atoi(optarg); //column with respect to which the data should be aligned
		      break;
	      }
	      case ('p'):			//print erromessages or warninngs or both
	      {
		      switch(optarg[0])
		      {
			      case 'a':
				      OPTIONS[11] = 1;
				      OPTIONS[12] = 1;
				      break;
			      case 'w':
				      OPTIONS[11] = 1;	//output error mesages
				      break;
			      case 'e':
				      OPTIONS[12] = 1;	//output warnnings
				      break;   
		      }
		      break;
	      }
	      case ('v'):
	      {
		      if(strcmp(optarg, "1.0") == 0){
		      	OPTIONS[0] = 1;
		      }
		      else{
		      	OPTIONS[0] = 2;
		      }
		      break;
	      }
	      case '?':
	      {
	        if (isprint (optopt)){
	          fprintf (stderr, "Unknown option `-%c'.\n", optopt);
		}
	        else{
	          fprintf (stderr, "Unknown option character `\\x%x'.\n", optopt);
		}
		printHelp();
	        exit (1);
	      }
	      default:
	      {
		printHelp();
	        exit(1);
	      }
	   }	
/////////////CHECK FOR VALIDITY IF THE GUI IS NOT REQUESTED///////////////////////////////////
	if(ret != 2){
		//check whether the input file exists////////////////////////////
		if( (inputfile[0] == '\0') || (strcmp(inputfile, "-") == 0) ){		//is this stdin
					
		}
		else{
			FILE *tmpfile;
			tmpfile = fopen(inputfile, "r");
			if(!tmpfile){
				fprintf(stderr, "the file %s does not exist\n", inputfile); 
			return 0;
			}
			else{
				fclose(tmpfile);
			}
		}
		*pinfile = inputfile;
		//check the outputfile name
		if( (outputfile[0] == '\0') || (strcmp(outputfile, "-") == 0) ){
			//output is stndard output 
			OUTF = stdout;
		}
		else{
			OUTF = fopen(outputfile, "w");
			if(!OUTF){
				fprintf(stderr, "there is a problem opening the file %s \n", outputfile);
				return 0;
			}
		}
		//check is folding or unfolding set
		if(OPTIONS[1] == -1){
			fprintf(stderr, 
			"you should use either -w or -u to specify wrap/folding or unwrap/unfolding. Program aborted\n");
			return 0;
		}
		//if format everything except comments and format only comments is set unset both
		  if((OPTIONS[7] == 1) && (OPTIONS[8] == 1)){
		  	OPTIONS[7] = 0;
			OPTIONS[8] = 0;
		  }
		  //if column with respect to which the data should be left justified is less than 5 
		  // or bigger than maxlen - 10 warn the user and reset it to 0
		  if( (OPTIONS[9] != -1) && ((OPTIONS[9] < 5) || (OPTIONS[9] > (MAXLENGTH - 10))) ){
		  	fprintf(stderr, "invalid use of the option -r usage:\n");
			fprintf(stderr, "-r takes int > 5 and < maxlen -10\n");
			OPTIONS[9] = 0;		//reset it
		  }
		if( (MAXLENGTH < 60) ||( MAXLENGTH > 2048)){
			fprintf(stderr, "invalid use of option -m usage:\n");
			fprintf(stderr, "-m takes int >= 60 and <= 2048\n");
			  if(MAXLENGTH < 60){
			  	MAXLENGTH = 60;
				fprintf(stderr, "MAXLENGTH set to 60\n");
			  }
			  else{
			  	MAXLENGTH = 2048;
				fprintf(stderr, "MAXLENGTH set to 2048\n");
		  	}
	  	}
		//print the no arguments if any	  
		for (index = optind; index < argc; index++){
		    fprintf (stderr, "Non-option argument %s\n", argv[index]);
	  	}
	}//if ret != 2
	return ret;
	}

	int tokanize (char *totokanize){
		char **ptoken;
		char *token = new char[300];
		char Delim[] = "-";	
		char Delimp[] = ",";
		unsigned long  start = 0; //the first int
		unsigned long  end = 0;	  //the second int
		int count = 0;		  //the first or the second int
		int error = 0;		  //the number of errors during the conversion
		char *wrong = NULL;	  //if there is something stored an error occured (see man strtol())
		char *tok = new char[300]; 		
		char **ptok;
	       	
		*ptoken = token;
		*ptok = tok;
		
		tok = strtok_r(totokanize, Delimp, ptok );
		while(tok != NULL){
			token = strtok_r(tok, Delim, ptoken);
			while(token != NULL)
			{
				if(count == 1){
					
					end = strtol(token, &wrong, 10);
					count = 0;
					if(*wrong != '\0'){
						error++;
						break;
					}
					if( (start >= end) ||  ((!FileChunks.empty())&& (start <= FileChunks.back())) ){
						error ++;
						break;
					}
					else{
//	printf("%d  %d, ", start, end);					
						FileChunks.push_back(start);
						FileChunks.push_back(end);
					}
				}
				else{
					start = strtol(token, &wrong, 10);
					if(*wrong != '\0'){
						error++;
						break;
					}
					count++;
				}
				token = strtok_r(NULL, Delim, ptoken);
			}
			tok = strtok_r(NULL, Delimp, ptok);
		}
		if (error){
			fprintf(stderr, "invalid use of option -h usage:\n");
			fprintf(stderr, "input must be in pairs n-n,n-n where n is an int, the first int should be smaller than the second and bigger than any previously entered\n");
			return 0;
		}
		return 1;
	}
	void printHelp(){
		cerr<<endl;
		cerr<<"Usage:"<<endl;
		cerr<<"     ciffold [-i input_cif] [-o output_cif] [-x n-n,n-n]"<<endl;
		cerr<<"     [-l n] [-m n] [-C n] [-p a[w][e]] [-v file_vers]"<<endl;
		cerr<<"     [-c] [-d] [-e] [-g] [-w [-n]] [-u] [-L] [-t] [-h] [-M] [-V]"<<endl;
		cerr<<endl;
		cerr<<"input_cif defaults to stdin"<<endl;
		cerr<<"output_cif defaults to stdout"<<endl;
		cerr<<"input_cif of \"-\" is stdin, output_cif of \"-\" is stdout"<<endl;
		cerr<<endl;
		cerr<<"-x has integer values (see the manual for examples)"<<endl;
		cerr<<"-l has integer values"<<endl;
		cerr<<"-m has integer values from 60 to 2048 for a maximum line length"<<endl;
		cerr<<"-C has integer values (e.g 60 the date will be aligned with respect to"<<endl;
		cerr<<"column 60)"<<endl;
		cerr<<"-p has character values:"<<endl;
		cerr<<"	a - warnings and error messages will be printed"<<endl;
		cerr<<"	w - warnings"<<endl;
		cerr<<"	e - error messages"<<endl;
		cerr<<"-v has string values 1.0 or 1.1"<<endl;
		cerr<<"-c format only the comments of input_cif"<<endl;
		cerr<<"-d indicate that input_cif is a dictionary file"<<endl;
		cerr<<"-e format everything except comments"<<endl;
		cerr<<"-g invoke the GUI interface"<<endl;
		cerr<<"-w wrap/fold input_cif"<<endl;
		cerr<<"-n minimal folding"<<endl;
		cerr<<"-u unwrap/unfold input_cif"<<endl;
		cerr<<"-L preserve leading blanks"<<endl;
		cerr<<"-t format input_cif tersely"<<endl;
		cerr<<"-h print this help message"<<endl;
		cerr<<"-M use/generate a map"<<endl;
		cerr<<"-V prints the ciffold version"<<endl;
	}
	
